//
//  SynchronizationServer.m
//  TicTacToe
//
//  Created by Ingo Kasprzak on 11.10.12.
//  Copyright (c) 2012 Silutions. All rights reserved.
//

#import "SynchronizationServer.h"
#import "Device+Methods.h"

// Keys for synchronization dictionary
#define SyncDatabaseSyncCommandKey @"command"
#define SyncDatabaseSyncDataKey @"data"
#define SyncDatabaseSyncDateKey @"date"
#define SyncDatabaseSyncSenderUIDKey @"senderUID"
#define SyncDatabaseSyncDataUIDKey @"dataUID"
#define SyncDatabaseSyncWantsReceiptKey @"wantsReceipt"


@interface SynchronizationServer()
@property (strong) NSMutableDictionary *deviceUIDsByPeerID;
@property (strong) NSMutableDictionary *peerIDsBydeviceUID;

@property (strong) NSMutableDictionary *unconfirmedCommands;
@end


@implementation SynchronizationServer


#pragma mark - Public methods

- (NSArray*)connectedDeviceUIDs {
	return [self.deviceUIDsByPeerID allValues];
}

- (NSString*)deviceUIDForPeerID:(NSString*)peerID {
	return [self.deviceUIDsByPeerID objectForKey:peerID];
}

- (NSArray*)peerIdentifiersForDeviceUIDs:(NSArray*)deviceUIDs {
	NSArray *peerIDs = [self.peerIDsBydeviceUID objectsForKeys:deviceUIDs notFoundMarker:[NSNull null]];
	return [peerIDs filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"SELF!=NULL"]];
}

- (void)addConnectedDevice:(NSString*)deviceUID withPeerIdentifier:(NSString*)peerID {
	[self.deviceUIDsByPeerID setObject:deviceUID forKey:peerID];
	[self.peerIDsBydeviceUID setObject:peerID    forKey:deviceUID];
}

- (void)removePeerIdentifier:(NSString*)peerID {
	NSString *deviceUID = [self.deviceUIDsByPeerID objectForKey:peerID];
	[self.deviceUIDsByPeerID removeObjectForKey:peerID];
	if (deviceUID) {
		[self.peerIDsBydeviceUID removeObjectForKey:deviceUID];
	}
}

#pragma mark - Start & Stop

- (BOOL) startServer {
	self.unconfirmedCommands = [NSMutableDictionary dictionaryWithCapacity:2];
	self.deviceUIDsByPeerID = [NSMutableDictionary dictionaryWithCapacity:2];
	self.peerIDsBydeviceUID = [NSMutableDictionary dictionaryWithCapacity:2];
	return YES;
}

- (BOOL) stopServer {
	self.unconfirmedCommands = nil;
	self.deviceUIDsByPeerID = nil;
	self.peerIDsBydeviceUID = nil;
	return YES;
}

#pragma mark - Abstract methods

- (BOOL)sendData:(NSData*)data toDevices:(NSArray*)deviceUIDs {
	DBLog(@"Method not implemented!");
	return NO;
}

- (BOOL)sendData:(NSData*)data toPeers:(NSArray*)peerIDs {
	DBLog(@"Method not implemented!");
	return NO;
}

#pragma mark - Internal methods

- (NSDictionary*)dictionaryForCommand:(NSString*)command withObject:(id)object wantsReceipt:(BOOL)yesOrNo {
	// Eindeutige Datensatz-ID erstellen
	CFUUIDRef uuid = CFUUIDCreate(kCFAllocatorDefault);
	NSString* myUcid = (__bridge_transfer NSString*)CFUUIDCreateString(kCFAllocatorDefault,uuid);
	CFRelease(uuid);
    
    // Datenpaket erstellen
	NSDictionary *dict = @{
	SyncDatabaseSyncCommandKey		: command,
	SyncDatabaseSyncSenderUIDKey	: self.delegate.thisDevice.uid,
	SyncDatabaseSyncDataUIDKey		: myUcid,
	SyncDatabaseSyncDateKey			: [NSDate date],
	SyncDatabaseSyncWantsReceiptKey	: @(yesOrNo),
	SyncDatabaseSyncDataKey			: object
	};
	return dict;
}

- (BOOL)sendObject:(id)object withCommand:(NSString*)command toPeers:(NSArray*)peerIDs wantsReceipt:(BOOL)yesOrNo {
	NSDictionary *dict = [self dictionaryForCommand:(NSString*)command withObject:(id)object wantsReceipt:yesOrNo];
	[self sendData:[NSKeyedArchiver archivedDataWithRootObject:dict] toPeers:peerIDs];

	if (yesOrNo) {
		[self.unconfirmedCommands setObject: @{
				SyncDatabaseSyncCommandKey : command,
				   SyncDatabaseSyncDateKey : [NSDate date]}
									 forKey:[dict objectForKey:SyncDatabaseSyncDataUIDKey]];
	}
	return YES;
}

- (BOOL)sendObject:(id)object withCommand:(NSString*)command toDevices:(NSArray*)deviceUIDs wantsReceipt:(BOOL)yesOrNo {
	NSDictionary *dict = [self dictionaryForCommand:(NSString*)command withObject:(id)object wantsReceipt:yesOrNo];
	[self sendData:[NSKeyedArchiver archivedDataWithRootObject:dict] toDevices:deviceUIDs];
	
	if (yesOrNo) {
		[self.unconfirmedCommands setObject: @{
				SyncDatabaseSyncCommandKey : command,
				   SyncDatabaseSyncDateKey : [NSDate date]}
									 forKey:[dict objectForKey:SyncDatabaseSyncDataUIDKey]];
	}
	return YES;
}

- (BOOL)receivedData:(NSData*)data fromPeer:(NSString*)peer {
	NSDictionary *dict = [NSKeyedUnarchiver unarchiveObjectWithData:data];
	NSString *command = [dict objectForKey:SyncDatabaseSyncCommandKey];
	
	if ([command isEqualToString:SyncDatabaseSyncSynchronizeCommand]) {
		DBLog(@"Received synchronization data from device '%@'", [dict objectForKey:SyncDatabaseSyncSenderUIDKey]);
		[self receivedModifiedObjects:[dict objectForKey:SyncDatabaseSyncDataKey]];

	} else if ([command isEqualToString:SyncDatabaseSyncSynchronizeBidirectionalCommand]) {
		DBLog(@"Received synchronization data (bidirectional request) from device '%@'", [dict objectForKey:SyncDatabaseSyncSenderUIDKey]);
		[self receivedBidirectionalRequestAndModifiedObjects:[dict objectForKey:SyncDatabaseSyncDataKey]];

	} else if ([command isEqualToString:SyncDatabaseSyncReceiptCommand]) {
		DBLog(@"Received receipt from device '%@'", [dict objectForKey:SyncDatabaseSyncSenderUIDKey]);
		[self receivedReceiptForCommand:[dict objectForKey:SyncDatabaseSyncDataKey] fromPeer:peer];

	} else if ([command isEqualToString:SyncDatabaseSyncIdentificationCommand]) {
		DBLog(@"Identification sent from device '%@'", [dict objectForKey:SyncDatabaseSyncSenderUIDKey]);
		[self receivedIdentification:[dict objectForKey:SyncDatabaseSyncDataKey] fromPeer:peer];
	}
	
	// wenn vom Sender ein receipt angefordert wurde, receipt absenden
	if ([[dict objectForKey:SyncDatabaseSyncWantsReceiptKey] boolValue]) {
		[self sendObject:[dict objectForKey:SyncDatabaseSyncDataUIDKey]
			 withCommand:SyncDatabaseSyncReceiptCommand
			   toDevices:@[ [dict objectForKey:SyncDatabaseSyncSenderUIDKey] ]
			wantsReceipt:NO];
	}
	return YES;
}


- (void)receivedModifiedObjects:(NSSet*)objects {
	[self.delegate receivedModifiedObjects:objects];
}

- (void)receivedBidirectionalRequestAndModifiedObjects:(NSSet*)objects {
	[self.delegate receivedBidirectionalRequestAndModifiedObjects:objects];
}

- (void)receivedReceiptForCommand:(NSString*)ucid fromPeer:(NSString*)peer {
	if ([self.delegate respondsToSelector:@selector(receivedReceiptForCommand:fromDevice:withDate:)]) {
		NSDictionary *dict = [self.unconfirmedCommands objectForKey:ucid];
		NSString *deviceUID = [self.deviceUIDsByPeerID objectForKey:peer];
		
		[self.delegate receivedReceiptForCommand:[dict objectForKey:SyncDatabaseSyncCommandKey]
									  fromDevice:deviceUID
										withDate:[dict objectForKey:SyncDatabaseSyncDateKey]];
	}
    
	[self.unconfirmedCommands removeObjectForKey:ucid];
}

- (void)receivedIdentification:(NSString*)deviceUID fromPeer:(NSString*)peer {
	if ([self.delegate respondsToSelector:@selector(receivedIdentification:fromPeer:)]) {
		[self.delegate receivedIdentification:deviceUID fromPeer:peer];
	}
}

- (void)receivedDisconnection:(NSString*)deviceUID fromPeer:(NSString*)peer {
	if ([self.delegate respondsToSelector:@selector(receivedDisconnection:fromPeer:)]) {
		[self.delegate receivedDisconnection:deviceUID fromPeer:peer];
	}
}

@end
